from pathlib import Path
from typing import Tuple, Dict, Union, Optional

import numpy as np
import logging
from os import path
from PIL import Image

from .vendor.dm3 import DM3
from .vendor.dm4 import DM4File, DM4TagHeader
from tifffile import TiffFile
from czifile import CziFile

# Disable PIL's maximum image limit.
Image.MAX_IMAGE_PIXELS = None


class ImageReader:
    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        pass

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        pass

    def read_channel_count(self, file_name: Path) -> int:
        pass

    # the sample count describes how many values are saved for each pixel, e.g. uint24 is 3x uint8 channels, therefore the sample count is 3. This applies only on TIFF and CZI datasets.
    def read_sample_count(
        self, file_name: Path  # pylint: disable=unused-argument
    ) -> int:
        return 1

    def read_z_slices_per_file(
        self, file_name: Path  # pylint: disable=unused-argument
    ) -> int:
        return 1

    def read_dtype(self, file_name: Path) -> str:
        raise NotImplementedError()


class PillowImageReader(ImageReader):
    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        this_layer = np.array(Image.open(file_name), dtype)
        this_layer = this_layer.swapaxes(0, 1)
        if channel_index is not None and this_layer.ndim == 3:
            this_layer = this_layer[:, :, channel_index]
        this_layer = this_layer.reshape(this_layer.shape + (1,))
        return this_layer

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        with Image.open(file_name) as test_img:
            return test_img.width, test_img.height

    def read_channel_count(self, file_name: Path) -> int:
        with Image.open(file_name) as test_img:
            this_layer = np.array(test_img)
            if this_layer.ndim == 2:
                # For two-dimensional data, the channel count is one
                return 1
            else:
                return this_layer.shape[-1]  # pylint: disable=unsubscriptable-object

    def read_dtype(self, file_name: Path) -> str:
        return np.array(Image.open(file_name)).dtype.name


def to_target_datatype(data: np.ndarray, target_dtype: np.dtype) -> np.ndarray:
    factor = (1 + np.iinfo(data.dtype).max) / (1 + np.iinfo(target_dtype).max)
    return (data / factor).astype(target_dtype)


class Dm3ImageReader(ImageReader):
    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        dm3_file = DM3(file_name)
        this_layer = to_target_datatype(dm3_file.imagedata, dtype)
        this_layer = this_layer.swapaxes(0, 1)
        this_layer = this_layer.reshape(this_layer.shape + (1,))
        return this_layer

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        test_img = DM3(file_name)
        return test_img.width, test_img.height

    def read_channel_count(self, _file_name: Path) -> int:
        logging.info("Assuming single channel for DM3 data")
        return 1

    def read_dtype(self, file_name: Path) -> str:
        return DM3(file_name).imagedata.dtype.name


class Dm4ImageReader(ImageReader):
    def _read_tags(self, dm4file: DM4File) -> Tuple[DM4File.DM4TagDir, DM4TagHeader]:
        tags = dm4file.read_directory()
        image_data_tag = (
            tags.named_subdirs["ImageList"]
            .unnamed_subdirs[1]
            .named_subdirs["ImageData"]
        )
        image_tag = image_data_tag.named_tags["Data"]

        return image_data_tag, image_tag

    def _read_dimensions(
        self, dm4file: DM4File, image_data_tag: DM4File.DM4TagDir
    ) -> Tuple[int, int]:
        width = dm4file.read_tag_data(
            image_data_tag.named_subdirs["Dimensions"].unnamed_tags[0]
        )
        height = dm4file.read_tag_data(
            image_data_tag.named_subdirs["Dimensions"].unnamed_tags[1]
        )
        return width, height

    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        dm4file = DM4File.open(str(file_name))
        image_data_tag, image_tag = self._read_tags(dm4file)
        width, height = self._read_dimensions(dm4file, image_data_tag)

        data = np.array(dm4file.read_tag_data(image_tag), dtype=np.uint16)

        data = data.reshape((width, height)).T
        data = np.expand_dims(data, 2)
        data = to_target_datatype(data, dtype)

        dm4file.close()

        return data

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        dm4file = DM4File.open(file_name)
        image_data_tag, _ = self._read_tags(dm4file)
        dimensions = self._read_dimensions(dm4file, image_data_tag)
        dm4file.close()

        return dimensions

    def read_channel_count(self, _file_name: Path) -> int:
        logging.info("Assuming single channel for DM4 data")
        return 1

    def read_dtype(self, file_name: Path) -> str:  # pylint: disable=unused-argument
        # DM4 standard input type is uint16
        return "uint16"


class TiffImageReader(ImageReader):
    @staticmethod
    def _get_page_data(tif_file: TiffFile, page_index: int) -> np.ndarray:
        if 0 <= page_index < len(tif_file.pages):
            return tif_file.pages[page_index].asarray()
        elif tif_file.is_imagej:
            # ImageJ can export 3-dimensional, single-page tiff files.
            # In that case, we manually read from the correct filehandle
            # by seeking to the correct position.
            shape = tif_file.pages[0].shape
            elements_per_page = 1
            for i in shape:
                elements_per_page *= i
            series = tif_file.series[0]
            tif_file.filehandle.seek(
                series.offset
                + page_index * np.dtype(series.dtype).itemsize * elements_per_page
            )
            data = tif_file.filehandle.read_array(series.dtype, elements_per_page)
            data = data.reshape(shape)
            return data
        else:
            raise Exception("Cannot read page data!")

    @staticmethod
    def _get_page_index(
        tif_file: TiffFile,
        z_index: int,
        c_index: int,
        s_index: int,
        axes: Dict[str, Tuple[bool, int, int]],
        z_axis_name: str,
        page_axes: Dict[str, Tuple[int, int, int]],
    ) -> int:
        # TiffFile consists of Tiff Pages, which are ordered according to the axis order
        # To find the index of the page that contains the data we want, we increase the page_index for each axis that we are looking for
        # Additionally, we multiply the index with dimension of each axis to "skip" the necessary amount of pages.
        # We divide by an axis if it is present in the TiffPage because there are less TiffPages to skip.
        page_index = 0
        for axis in tif_file.series[0].axes:
            page_index *= axes[axis][2]
            if axis in page_axes:
                page_index //= page_axes[axis][2]
            if axis == z_axis_name:
                page_index += z_index
            elif axis == "C":
                page_index += c_index
            elif axis == "S":
                page_index += s_index
        return page_index

    @staticmethod
    def _find_correct_pages(
        tif_file: TiffFile,
        channel_index: Optional[int],
        sample_index: Optional[int],
        z_index: int,
        axes: Dict[str, Tuple[bool, int, int]],
        z_axis_name: str,
        page_axes: Dict[str, Tuple[int, int, int]],
    ) -> Dict[int, Tuple[Optional[int], Optional[int]]]:
        # return format is page_index to [Optional[channel_index], Optional[sample_index]]
        result: Dict[int, Tuple[Optional[int], Optional[int]]] = dict()

        if channel_index is not None:
            page_c_index = (
                (channel_index // page_axes["C"][2])
                if "C" in page_axes
                else channel_index
            )
            data_c_index = (
                (channel_index % page_axes["C"][2]) if "C" in page_axes else None
            )
            page_s_index = 0
            data_s_index = None
            if sample_index is not None:
                page_s_index = (
                    (sample_index // page_axes["S"][2])
                    if "S" in page_axes
                    else sample_index
                )
                data_s_index = (
                    (sample_index % page_axes["S"][2]) if "S" in page_axes else None
                )

            result[
                TiffImageReader._get_page_index(
                    tif_file,
                    z_index,
                    page_c_index,
                    page_s_index,
                    axes,
                    z_axis_name,
                    page_axes,
                )
            ] = (data_c_index, data_s_index)
        else:
            # implies sample_index not set
            c_range = range(1)
            s_range = range(1)
            if "C" in page_axes and axes["C"][0]:
                c_range = range(axes["C"][2] // page_axes["C"][2])
            elif axes["C"][0]:
                c_range = range(axes["C"][2])

            if "S" in page_axes and axes["S"][0]:
                s_range = range(axes["S"][2] // page_axes["S"][2])
            elif axes["S"][0]:
                s_range = range(axes["S"][2])

            for c in c_range:
                for s in s_range:
                    result[
                        TiffImageReader._get_page_index(
                            tif_file, z_index, c, s, axes, z_axis_name, page_axes
                        )
                    ] = (None, None)

        return result

    @staticmethod
    def find_count_of_axis(tif_file: TiffFile, axis: str) -> int:
        assert len(tif_file.series) == 1, "only single tif series are supported"
        tif_series = tif_file.series[0]
        index = tif_series.axes.find(axis)
        if index == -1:
            return 1
        else:
            return tif_series.shape[index]  # pylint: disable=unsubscriptable-object

    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        with TiffFile(file_name) as tif_file:
            # Read metadata
            z_axis_name = self._find_right_z_axis_name(tif_file)
            used_axes = {"C", "Y", "X", "S"} | {z_axis_name}
            assert len(tif_file.series) == 1, "Multi-series tiff not supported"
            series = tif_file.series[0]  # pylint: disable=unsubscriptable-object
            page = tif_file.pages[0]
            axes = {}
            for index, axis in enumerate(series.axes):
                axes[axis] = (True, index, series.shape[index])
            for axis in used_axes:
                if axis not in axes:
                    axes[axis] = (False, 0, 1)
            count = 0
            page_axes = {}
            for index, axis in enumerate(page.axes):
                # when reading the data, we'll remove any axis that is not used, so the dimensionality shrinks
                page_axes[axis] = (index, index - count, page.shape[index])
                if axis not in used_axes:
                    count += 1
            assert (
                len(axes) > 0 and z_axis_name is not None
            ), "TiffReader initialization failed"
            # this calculation is possible this way because the extent for non existing axes gets set to one
            num_output_channel = axes["C"][2] * axes["S"][2]
            if channel_index is not None:
                num_output_channel //= axes["C"][2]
            if sample_index is not None:
                num_output_channel //= axes["S"][2]

            output_shape = (axes["X"][2], axes["Y"][2], num_output_channel)
            output = np.empty(output_shape, tif_file.pages[0].dtype)

            z_index = z_slice if axes[z_axis_name][0] else 0
            output_channel_offset = 0

            # Read the data
            for page_index, (data_c_index, data_s_index,) in self._find_correct_pages(
                tif_file,
                channel_index,
                sample_index,
                z_index,
                axes,
                z_axis_name,
                page_axes,
            ).items():
                page_data = self._get_page_data(tif_file, page_index)
                # remove any axis that we do not use
                # left over axes [(C)YX(S)] (Z axis is not on same TiffPage)
                for axis in page_axes:
                    if axis not in used_axes:
                        page_data = page_data.take([0], page_axes[axis][1])

                # axes order is then [YX(S)(C)]
                if page_data.ndim == 4:
                    page_data = page_data.transpose(
                        (
                            page_axes["X"][1],
                            page_axes["Y"][1],
                            page_axes["S"][1],
                            page_axes["C"][1],
                        )
                    )
                elif page_data.ndim == 3:
                    if "S" in page_axes:
                        page_data = page_data.transpose(
                            (
                                page_axes["X"][1],
                                page_axes["Y"][1],
                                page_axes["S"][1],
                            )
                        )
                    else:
                        page_data = page_data.transpose(
                            (
                                page_axes["X"][1],
                                page_axes["Y"][1],
                                page_axes["C"][1],
                            )
                        )
                elif page_data.ndim == 2:
                    page_data = page_data.transpose(
                        (page_axes["X"][1], page_axes["Y"][1])
                    )
                else:
                    raise Exception

                if data_c_index is not None:
                    page_data = page_data.take([data_c_index], -1)
                if data_s_index is not None:
                    page_data = page_data.take([data_s_index], -1)

                if page_data.ndim == 4:
                    # means C and S axis present and no data index set
                    for c in range(page_data.shape[-1]):
                        output[
                            :,
                            :,
                            output_channel_offset : output_channel_offset
                            + page_axes["S"][2],
                        ] = page_data[:, :, :, c]
                        output_channel_offset += (
                            output_channel_offset + page_axes["S"][2]
                        )
                elif page_data.ndim == 3:
                    next_channel_offset = output_channel_offset + page_data.shape[-1]
                    output[:, :, output_channel_offset:next_channel_offset] = page_data
                    output_channel_offset = next_channel_offset
                elif page_data.ndim == 2:
                    output[:, :, output_channel_offset] = page_data
                    output_channel_offset += 1
                else:
                    raise Exception("Invalid axis count")

            output = np.array(output, dtype)
            output = output.reshape(output.shape + (1,))
            return output

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        with TiffFile(file_name) as tif_file:
            return (
                TiffImageReader.find_count_of_axis(tif_file, "X"),
                TiffImageReader.find_count_of_axis(tif_file, "Y"),
            )

    def read_channel_count(self, file_name: Path) -> int:
        with TiffFile(file_name) as tif_file:
            return TiffImageReader.find_count_of_axis(tif_file, "C")

    def read_sample_count(self, file_name: Path) -> int:
        with TiffFile(file_name) as tif_file:
            return TiffImageReader.find_count_of_axis(tif_file, "S")

    def read_z_slices_per_file(self, file_name: Path) -> int:
        with TiffFile(file_name) as tif_file:
            z_axis_name = self._find_right_z_axis_name(tif_file)
            return TiffImageReader.find_count_of_axis(tif_file, z_axis_name)

    def read_dtype(self, file_name: Path) -> str:
        with TiffFile(file_name) as tif_file:
            return tif_file.series[  # pylint: disable=unsubscriptable-object
                0
            ].dtype.name

    def _find_right_z_axis_name(self, tif_file: TiffFile) -> str:
        i_count = TiffImageReader.find_count_of_axis(tif_file, "I")
        z_count = TiffImageReader.find_count_of_axis(tif_file, "Z")
        q_count = TiffImageReader.find_count_of_axis(tif_file, "Q")
        if i_count > 1:
            assert (
                z_count * q_count == 1
            ), "Format error, as multiple Z axis names were identified"
            return "I"
        elif q_count > 1:
            assert (
                z_count * i_count == 1
            ), "Format error, as multiple Z axis names were identified"
            return "Q"
        else:
            assert (
                i_count * q_count == 1
            ), "Format error, as multiple Z axis names were identified"
            return "Z"


class CziImageReader(ImageReader):
    @staticmethod
    def _select_correct_tiles_for_czi(
        channel_index: Optional[int],
        sample_index: Optional[int],
        z_slice: int,
        axes: Dict[str, Tuple[bool, int, int]],
        tile_shape: Tuple[int, int],
        dataset_shape: Tuple[int, ...],
    ) -> Dict[int, Tuple[int, Optional[int], Optional[int], Optional[int]]]:
        assert len(axes) != 0, "CZIReader initialization failed"
        # return type is tile channel index to tile z index, data z index, data channel index and 0 index, None means all
        result: Dict[
            int, Tuple[int, Optional[int], Optional[int], Optional[int]]
        ] = dict()
        # C axis exists
        c_count_per_tile = 1
        c_tiles_for_complete_axis = 1
        if axes["C"][0]:
            c_count_per_tile = tile_shape[axes["C"][1]]
            c_tiles_for_complete_axis = dataset_shape[axes["C"][1]] // c_count_per_tile

        tile_z_index: int = 0
        data_z_index: Optional[int] = None
        if axes["Z"][0]:
            tile_z_index = z_slice // dataset_shape[axes["Z"][1]]
            data_z_index = z_slice % dataset_shape[axes["Z"][1]]

        if channel_index is not None:
            tile_channel_index = channel_index // c_count_per_tile
            data_channel_index = channel_index % c_count_per_tile
            result[tile_channel_index] = (
                tile_z_index,
                data_z_index,
                data_channel_index,
                sample_index,
            )
        else:
            # no channel index set, implies no sample index set
            for i in range(c_tiles_for_complete_axis):
                result[i] = (tile_z_index, data_z_index, None, None)

        return result

    @staticmethod
    def _matches_if_exist(
        axis: str, value: int, other_value: int, axes: Dict[str, Tuple[bool, int, int]]
    ) -> bool:
        assert len(axes) != 0, "Axes initialization failed"
        if axes[axis][0]:
            return value == other_value
        else:
            return True

    @staticmethod
    def find_count_of_axis(czi_file: CziFile, axis: str) -> int:
        index = czi_file.axes.find(axis)
        if index == -1:
            return 1
        else:
            return czi_file.shape[index]

    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int],
        sample_index: Optional[int],
    ) -> np.ndarray:
        with CziFile(file_name) as czi_file:
            # Read metadata
            tile_shape = czi_file.filtered_subblock_directory[  # pylint: disable=unsubscriptable-object
                0
            ].shape
            dataset_shape: Tuple[int, ...] = tuple(czi_file.shape)
            count = 0
            used_axes = {"Z", "C", "Y", "X", "0"}
            axes: Dict[str, Tuple[bool, int, int]] = {}
            for i, axis in enumerate(czi_file.axes):
                axes[axis] = (True, i, i - count)
                if axis not in used_axes:
                    count += 1
            for axis in used_axes:
                if axis not in axes:
                    axes[axis] = (False, 0, 0)

            assert tile_shape is not None, "Cannot read tile shape format."

            if sample_index is not None:
                num_output_channel = 1
            elif channel_index is not None:
                num_output_channel = dataset_shape[axes["0"][1]]
            else:
                channel_count = dataset_shape[axes["C"][1]] if axes["C"][0] else 1
                num_output_channel = channel_count * dataset_shape[axes["0"][1]]

            output_shape = (
                dataset_shape[axes["X"][1]],
                dataset_shape[axes["Y"][1]],
                num_output_channel,
            )
            output = np.empty(output_shape, dtype)

            z_file_start = (
                czi_file.start[axes["Z"][1]]  # pylint: disable=unsubscriptable-object
                if axes["Z"][0]
                else 0
            )
            c_file_start = (
                czi_file.start[axes["C"][1]]  # pylint: disable=unsubscriptable-object
                if axes["C"][0]
                else 0
            )
            output_channel_offset = 0

            # Read the data
            for tile_channel_index, (
                tile_z_index,
                data_z_index,
                data_channel_index,
                data_sample_index,
            ) in CziImageReader._select_correct_tiles_for_czi(
                channel_index, sample_index, z_slice, axes, tile_shape, dataset_shape
            ).items():
                # since the czi tiles are not sorted, we search linearly through them and check if the tile matches with the wanted coordinates
                # however, some axes might not exist, so the _matches_if_exist helper returns true if the axis does not exist
                for (
                    entry
                ) in (
                    czi_file.filtered_subblock_directory  # pylint: disable=not-an-iterable
                ):
                    if CziImageReader._matches_if_exist(
                        "Z",
                        entry.start[axes["Z"][1]] - z_file_start,
                        tile_z_index,
                        axes,
                    ) and CziImageReader._matches_if_exist(
                        "C",
                        (entry.start[axes["C"][1]] - c_file_start)
                        // tile_shape[axes["C"][1]],
                        tile_channel_index,
                        axes,
                    ):
                        data = entry.data_segment().data()
                        data = to_target_datatype(data, dtype)

                        # left axis are in order [(Z)(C)0XY]
                        for axis in czi_file.axes:  # pylint: disable=not-an-iterable
                            if axis not in used_axes:
                                data = data.take([0], axes[axis][2])
                        if data.ndim == 5:
                            data = data.transpose(
                                (
                                    axes["Z"][2],
                                    axes["X"][2],
                                    axes["Y"][2],
                                    axes["0"][2],
                                    axes["C"][2],
                                )
                            )
                        elif data.ndim == 4:
                            if axes["Z"][0]:
                                data = data.transpose(
                                    (
                                        axes["Z"][2],
                                        axes["X"][2],
                                        axes["Y"][2],
                                        axes["0"][2],
                                    )
                                )
                            else:
                                data = data.transpose(
                                    (
                                        axes["X"][2],
                                        axes["Y"][2],
                                        axes["0"][2],
                                        axes["C"][2],
                                    )
                                )
                        elif data.ndim == 3:
                            data = data.transpose(
                                (
                                    axes["X"][2],
                                    axes["Y"][2],
                                    axes["0"][2],
                                )
                            )
                        else:
                            raise Exception

                        if data_z_index is not None:
                            data = data.take([data_z_index], 0)

                        # if z axis exist, data_z_index is set, so the access index for this is always 0
                        if data_channel_index is not None and axes["C"][0]:
                            data = data.take([data_channel_index], -1)

                        # if the sample index is set, the channel index is set, too, so access index is always 0
                        if data_sample_index is not None:
                            data = data.take([data_sample_index], -1)

                        # special case with C and 0 and no selected index set
                        if data.ndim == 4:
                            for c in range(tile_shape[axes["C"][1]]):
                                next_output_channel_offset = (
                                    output_channel_offset + data.shape[2]
                                )
                                output[
                                    :,
                                    :,
                                    output_channel_offset:next_output_channel_offset,
                                ] = data[:, :, :, c]
                                output_channel_offset = next_output_channel_offset
                        elif data.ndim == 2:
                            output[:, :, output_channel_offset] = data
                            output_channel_offset += 1
                        else:
                            output_channel_increment = (
                                1 if data_sample_index is not None else data.shape[-1]
                            )
                            next_output_channel_offset = (
                                output_channel_offset + output_channel_increment
                            )
                            output[
                                :, :, output_channel_offset:next_output_channel_offset
                            ] = data
                            output_channel_offset = next_output_channel_offset

        # CZI stores pixels as BGR instead of RGB, so swap axes to ensure right color output
        if (
            num_output_channel == 3
            and dtype == np.uint8
            and czi_file.filtered_subblock_directory[  # pylint: disable=unsubscriptable-object
                0
            ].pixel_type
            == "Bgr24"
        ):
            output[:, :, 2], output[:, :, 0] = output[:, :, 0], output[:, :, 2]

        output = output.reshape(output.shape + (1,))

        return output

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        with CziFile(file_name) as czi_file:
            return (
                CziImageReader.find_count_of_axis(czi_file, "X"),
                CziImageReader.find_count_of_axis(czi_file, "Y"),
            )

    def read_channel_count(self, file_name: Path) -> int:
        with CziFile(file_name) as czi_file:
            return CziImageReader.find_count_of_axis(czi_file, "C")

    def read_sample_count(self, file_name: Path) -> int:
        with CziFile(file_name) as czi_file:
            return CziImageReader.find_count_of_axis(czi_file, "0")

    def read_z_slices_per_file(self, file_name: Path) -> int:
        with CziFile(file_name) as czi_file:
            return CziImageReader.find_count_of_axis(czi_file, "Z")

    def read_dtype(self, file_name: Path) -> str:
        with CziFile(file_name) as czi_file:
            return czi_file.dtype.name  # pylint: disable=no-member


class ImageReaderManager:
    def __init__(self) -> None:
        self.readers: Dict[
            str,
            Union[
                TiffImageReader,
                PillowImageReader,
                Dm3ImageReader,
                Dm4ImageReader,
                CziImageReader,
            ],
        ] = {
            # TIFF file endings
            ".tif": TiffImageReader(),
            ".tiff": TiffImageReader(),
            # JPEG file endings
            ".jpg": PillowImageReader(),
            ".jpeg": PillowImageReader(),
            ".jpe": PillowImageReader(),
            ".jfif": PillowImageReader(),
            ".jif": PillowImageReader(),
            # JPEG2000 file endings
            ".jp2": PillowImageReader(),
            ".j2k": PillowImageReader(),
            ".jpf": PillowImageReader(),
            ".jpm": PillowImageReader(),
            ".jpg2": PillowImageReader(),
            ".j2c": PillowImageReader(),
            ".jpc": PillowImageReader(),
            ".jpx": PillowImageReader(),
            ".mj2": PillowImageReader(),
            # Other image file endings
            ".png": PillowImageReader(),
            ".dm3": Dm3ImageReader(),
            ".dm4": Dm4ImageReader(),
            ".czi": CziImageReader(),
        }

    def read_array(
        self,
        file_name: Path,
        dtype: np.dtype,
        z_slice: int,
        channel_index: Optional[int] = None,
        sample_index: Optional[int] = None,
    ) -> np.ndarray:
        _, ext = path.splitext(file_name)

        # Image shape will be (x, y, channel_count, z=1) or (x, y, z=1)
        image = self.readers[ext].read_array(
            file_name, dtype, z_slice, channel_index, sample_index
        )
        # Standardize the image shape to (x, y, channel_count, z=1)
        if image.ndim == 3:
            image = image.reshape(image.shape + (1,))

        return image

    def read_dimensions(self, file_name: Path) -> Tuple[int, int]:
        _, ext = path.splitext(file_name)
        return self.readers[ext].read_dimensions(file_name)

    def read_channel_count(self, file_name: Path) -> int:
        _, ext = path.splitext(file_name)
        return self.readers[ext].read_channel_count(file_name)

    def read_sample_count(self, file_name: Path) -> int:
        _, ext = path.splitext(file_name)
        return self.readers[ext].read_sample_count(file_name)

    def read_z_slices_per_file(self, file_name: Path) -> int:
        _, ext = path.splitext(file_name)
        return self.readers[ext].read_z_slices_per_file(file_name)

    def read_dtype(self, file_name: Path) -> str:
        _, ext = path.splitext(file_name)
        return self.readers[ext].read_dtype(file_name)


image_reader = ImageReaderManager()



from argparse import ArgumentParser, Namespace
import os
from pathlib import Path

from typing import Optional

from wkcuber.api.dataset import Dataset
from wkcuber.downsampling_utils import SamplingModes
from .mag import Mag

from .utils import (
    add_verbose_flag,
    add_distribution_flags,
    add_interpolation_flag,
    add_isotropic_flag,
    setup_logging,
    add_sampling_mode_flag,
)


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument("path", help="Directory containing the dataset.", type=Path)

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--from_mag",
        "--from",
        "-f",
        help="Resolution to base upsampling on",
        type=str,
    )

    # Either provide the maximum resolution to be upsampled OR a specific, anisotropic magnification.
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--target_mag",
        "-t",
        help="Min resolution to be upsampled. In case of anisotropic upsampling, the process is considered "
        "done when max(current_mag) < max(target_mag) where max takes the largest dimension of the mag tuple "
        "x, y, z (or after Mag(1) was upsampled). For example, a minimum mag value of 2 (or 2-2-2) will stop "
        "the upsampling as soon as a magnification is produced for which one dimension is smaller than 2. "
        "The default value is Mag(1)",
        type=Mag,
        default=Mag(1),
    )

    group.add_argument(
        "--anisotropic_target_mag",
        help="Specify an explicit anisotropic target magnification (e.g., --anisotropic_target_mag 2-2-1)."
        "All magnifications until this target magnification will be created. Consider using --anisotropic "
        "instead which automatically creates multiple anisotropic magnifications depending "
        "on the dataset's scale",
        type=str,
    )

    parser.add_argument(
        "--buffer_cube_size",
        "-b",
        help="Size of buffered cube to be upsampled (i.e. buffer cube edge length)",
        type=int,
        default=None,
    )

    parser.add_argument(
        "--no_compress",
        help="Don't compress data during upsampling",
        default=False,
        action="store_true",
    )

    add_interpolation_flag(parser)
    add_verbose_flag(parser)
    add_isotropic_flag(parser)
    add_sampling_mode_flag(parser)
    add_distribution_flags(parser)

    return parser


def upsample_mags(
    path: Path,
    layer_name: Optional[str] = None,
    from_mag: Optional[Mag] = None,
    target_mag: Mag = Mag(1),
    buffer_edge_len: Optional[int] = None,
    compress: bool = True,
    args: Optional[Namespace] = None,
    sampling_mode: str = SamplingModes.ANISOTROPIC,
) -> None:
    assert layer_name and from_mag or not layer_name and not from_mag, (
        "You provided only one of the following "
        "parameters: layer_name, from_mag but both "
        "need to be set or none. If you don't provide "
        "the parameters you need to provide the path "
        "argument with the mag and layer to upsample"
        " (e.g dataset/color/1)."
    )
    if not layer_name or not from_mag:
        layer_name = os.path.basename(os.path.dirname(path))
        from_mag = Mag(os.path.basename(path))
        path = path.parent.parent

    Dataset.open(path).get_layer(layer_name).upsample(
        from_mag=from_mag,
        min_mag=target_mag,
        compress=compress,
        sampling_mode=sampling_mode,
        buffer_edge_len=buffer_edge_len,
        args=args,
    )


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    if args.isotropic is not None:
        raise DeprecationWarning(
            "The flag 'isotropic' is deprecated. Consider using '--sampling_mode isotropic' instead."
        )

    if args.anisotropic_target_mag is not None:
        raise DeprecationWarning(
            "The 'anisotropic_target_mag' flag is deprecated. Use '--max' instead (and consider changing the 'sampling_mode')"
        )

    from_mag = Mag(args.from_mag)
    target_mag = Mag(args.target_mag)

    upsample_mags(
        args.path,
        args.layer_name,
        from_mag,
        target_mag,
        buffer_edge_len=args.buffer_cube_size,
        compress=not args.no_compress,
        sampling_mode=args.sampling_mode,
        args=args,
    )



import logging
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Callable, Optional

from wkcuber.api.dataset import Dataset
from webknossos.geometry import BoundingBox, Vec3Int
import numpy as np

from wkcuber.mag import Mag
from .utils import (
    add_verbose_flag,
    open_wkw,
    WkwDatasetInfo,
    add_distribution_flags,
    get_executor_for_args,
    wait_and_ensure_success,
    setup_logging,
)
from .metadata import detect_resolutions, detect_bbox, detect_layers
import functools
from .compress import BACKUP_EXT


CHUNK_SIZE = Vec3Int.full(1024)


def named_partial(func: Callable, *args: Any, **kwargs: Any) -> Callable:
    # Propagate __name__ and __doc__ attributes to partial function
    partial_func = functools.partial(func, *args, **kwargs)
    functools.update_wrapper(partial_func, func)
    if hasattr(func, "__annotations__"):
        # Generic types cannot be pickled in Python <= 3.6, see https://github.com/python/typing/issues/511
        partial_func.__annotations__ = {}
    return partial_func


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument("source_path", help="Path to input WKW dataset", type=Path)

    parser.add_argument(
        "target_path",
        help="WKW dataset with which to compare the input dataset.",
        type=Path,
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the layer to compare (if not provided, all layers are compared)",
        default=None,
    )

    add_verbose_flag(parser)
    add_distribution_flags(parser)

    return parser


def assert_equality_for_chunk(
    source_path: Path,
    target_path: Path,
    layer_name: str,
    mag: Mag,
    sub_box: BoundingBox,
) -> None:
    wk_dataset = Dataset.open(source_path)
    layer = wk_dataset.get_layer(layer_name)
    backup_wkw_info = WkwDatasetInfo(target_path, layer_name, mag, header=None)
    with open_wkw(backup_wkw_info) as backup_wkw:
        mag_view = layer.get_mag(mag)
        logging.info(f"Checking sub_box: {sub_box}")

        data = mag_view.read(sub_box.topleft, sub_box.size)
        backup_data = backup_wkw.read(sub_box.topleft, sub_box.size)
        assert np.all(
            data == backup_data
        ), f"Data differs in bounding box {sub_box} for layer {layer_name} with mag {mag}"


def check_equality(
    source_path: Path, target_path: Path, args: Optional[Namespace] = None
) -> None:

    logging.info(f"Comparing {source_path} with {target_path}")

    wk_src_dataset = Dataset.open(source_path)
    src_layer_names = list(wk_src_dataset.layers.keys())
    target_layer_names = [
        layer["name"] for layer in detect_layers(target_path, 0, False)
    ]
    assert set(src_layer_names) == set(
        target_layer_names
    ), f"The provided input datasets have different layers: {src_layer_names} != {target_layer_names}"

    existing_layer_names = src_layer_names

    if args is not None and args.layer_name is not None:
        assert (
            args.layer_name in existing_layer_names
        ), f"Provided layer {args.layer_name} does not exist in input dataset."
        existing_layer_names = [args.layer_name]

    for layer_name in existing_layer_names:

        logging.info(f"Checking layer_name: {layer_name}")

        source_mags = list(detect_resolutions(source_path, layer_name))
        target_mags = list(detect_resolutions(target_path, layer_name))
        source_mags.sort()
        target_mags.sort()
        mags = source_mags

        assert (
            source_mags == target_mags
        ), f"The mags between {source_path}/{layer_name} and {target_path}/{layer_name} are not equal: {source_mags} != {target_mags}"

        official_bbox = wk_src_dataset.get_layer(layer_name).bounding_box

        for mag in mags:
            inferred_src_bbox_dict = detect_bbox(source_path, layer_name, mag)
            inferred_target_bbox_dict = detect_bbox(target_path, layer_name, mag)
            assert inferred_src_bbox_dict is not None
            assert inferred_target_bbox_dict is not None
            inferred_src_bbox = BoundingBox.from_auto(inferred_src_bbox_dict)
            inferred_target_bbox = BoundingBox.from_auto(inferred_target_bbox_dict)

            bbox = inferred_src_bbox.extended_by(inferred_target_bbox).extended_by(
                official_bbox
            )
            logging.info(f"Start verification of {layer_name} in mag {mag} in {bbox}")

            with get_executor_for_args(args) as executor:
                boxes = list(bbox.chunk(CHUNK_SIZE, CHUNK_SIZE))
                assert_fn = named_partial(
                    assert_equality_for_chunk, source_path, target_path, layer_name, mag
                )

                wait_and_ensure_success(executor.map_to_futures(assert_fn, boxes))

    logging.info(
        f"The following datasets seem to be equal (with regard to the layers: {existing_layer_names}):"
    )
    logging.info(source_path)
    logging.info(target_path)


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    if args.target_path is None:
        target_path = args.source_path + BACKUP_EXT
    else:
        target_path = args.target_path
    check_equality(args.source_path, target_path, args)



import argparse
from functools import partial
import logging
from pathlib import Path
import time
from typing import Optional, Tuple, Union, cast
import numpy as np
import zarr

from webknossos.dataset import SegmentationLayer

from webknossos.dataset.defaults import DEFAULT_WKW_FILE_LEN
from webknossos import BoundingBox, Dataset, Mag, MagView
from webknossos.utils import get_executor_for_args, wait_and_ensure_success
from .utils import (
    add_distribution_flags,
    add_interpolation_flag,
    add_sampling_mode_flag,
    add_scale_flag,
    add_verbose_flag,
    DEFAULT_WKW_VOXELS_PER_BLOCK,
    get_executor_args,
    setup_logging,
)


logger = logging.getLogger(__name__)


def parse_flip_axes(flip_axes: str) -> Tuple[int, ...]:
    try:
        indices = tuple(int(x) for x in flip_axes.split(","))
    except Exception as e:
        raise argparse.ArgumentTypeError("The flip_axes could not be parsed") from e
    if [i for i in indices if i < 0 or i > 3]:
        raise argparse.ArgumentTypeError("The flip_axes contains out-of-bound values")
    return indices


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "source_path",
        help="Path to zarr file to convert",
        type=Path,
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated WKW dataset.", type=Path
    )

    add_scale_flag(parser, required=False)

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation).",
        default="color",
    )

    parser.add_argument(
        "--is_segmentation_layer",
        "-sl",
        help="Set whether this is a segmentation layer, defaulting to a color layer if nothing is set.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--flip_axes",
        help="The axes which should be flipped. "
        "Input format is a comma separated list of axis indices. "
        "For example, 1,2,3 will flip the x, y and z axes.",
        default=None,
        type=parse_flip_axes,
    )

    parser.add_argument(
        "--no_compress",
        help="Don't compress saved data",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--max_mag",
        "-m",
        help="Max resolution to be downsampled. Needs to be a power of 2. In case of anisotropic downsampling, "
        "the process is considered done when max(current_mag) >= max(max_mag) where max takes the "
        "largest dimension of the mag tuple x, y, z. For example, a maximum mag value of 8 (or 8-8-8) "
        "will stop the downsampling as soon as a magnification is produced for which one dimension is "
        "equal or larger than 8. "
        "The default value is calculated depending on the dataset size. In the lowest Mag, the size will be "
        "smaller than 100vx per dimension",
        type=int,
        default=None,
    )

    add_sampling_mode_flag(parser)
    add_interpolation_flag(parser)

    add_distribution_flags(parser)

    add_verbose_flag(parser)

    return parser


def _zarr_chunk_converter(
    bounding_box: BoundingBox,
    source_zarr_path: Path,
    target_mag_view: MagView,
    flip_axes: Optional[Union[int, Tuple[int, ...]]],
) -> int:
    logging.info(f"Conversion of {bounding_box.topleft}")

    slices = bounding_box.to_slices()
    zarr_file = zarr.open(source_zarr_path, "r")
    source_data = zarr_file[slices][None, ...]

    if flip_axes:
        source_data = np.flip(source_data, flip_axes)

    contiguous_chunk = source_data.copy(order="F")
    target_mag_view.write(contiguous_chunk, bounding_box.topleft)

    return source_data.max()


def convert_zarr(
    source_zarr_path: Path,
    target_path: Path,
    layer_name: str,
    is_segmentation_layer: bool = False,
    scale: Optional[Tuple[float, float, float]] = (1.0, 1.0, 1.0),
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
    compress: bool = True,
    file_len: int = DEFAULT_WKW_FILE_LEN,
    executor_args: Optional[argparse.Namespace] = None,
) -> MagView:
    ref_time = time.time()

    f = zarr.open(source_zarr_path, "r")
    input_dtype = f.dtype
    shape = f.shape

    if scale is None:
        scale = 1.0, 1.0, 1.0
    wk_ds = Dataset(target_path, scale=scale, exist_ok=True)
    wk_layer = wk_ds.get_or_add_layer(
        layer_name,
        "segmentation" if is_segmentation_layer else "color",
        dtype_per_layer=np.dtype(input_dtype),
        num_channels=1,
        largest_segment_id=0,
    )
    wk_layer.bounding_box = BoundingBox((0, 0, 0), shape)
    wk_mag = wk_layer.get_or_add_mag("1", file_len=file_len, compress=compress)

    # Parallel chunk conversion
    with get_executor_for_args(executor_args) as executor:
        largest_segment_id_per_chunk = wait_and_ensure_success(
            executor.map_to_futures(
                partial(
                    _zarr_chunk_converter,
                    source_zarr_path=source_zarr_path,
                    target_mag_view=wk_mag,
                    flip_axes=flip_axes,
                ),
                wk_layer.bounding_box.chunk(
                    chunk_size=(DEFAULT_WKW_VOXELS_PER_BLOCK * file_len,) * 3
                ),
            )
        )

    if is_segmentation_layer:
        largest_segment_id = int(max(largest_segment_id_per_chunk))
        cast(SegmentationLayer, wk_layer).largest_segment_id = largest_segment_id

    logger.debug(
        "Conversion of {} took {:.8f}s".format(source_zarr_path, time.time() - ref_time)
    )
    return wk_mag


def main(args: argparse.Namespace) -> None:
    source_path = args.source_path

    if not source_path.is_dir():
        logger.error("source_path is not a directory")
        return

    executor_args = get_executor_args(args)

    mag_view = convert_zarr(
        source_path,
        args.target_path,
        layer_name=args.layer_name,
        is_segmentation_layer=args.is_segmentation_layer,
        scale=args.scale,
        flip_axes=args.flip_axes,
        compress=not args.no_compress,
        executor_args=executor_args,
    )

    mag_view.layer.downsample(
        from_mag=mag_view.mag,
        max_mag=None if args.max_mag is None else Mag(args.max_mag),
        interpolation_mode=args.interpolation_mode,
        compress=not args.no_compress,
        sampling_mode=args.sampling_mode,
        args=executor_args,
    )


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    main(args)



"""
# Dataset API

Please use the Dataset API at `webknossos.dataset`.
"""

from .bounding_box import BoundingBox
from .dataset import Dataset
from .layer import Layer
from .mag_view import MagView
from .view import View



from pathlib import Path
from argparse import ArgumentParser
from typing import Optional
from os import environ

from webknossos import webknossos_context
from webknossos.client._defaults import DEFAULT_WEBKNOSSOS_URL
from webknossos.client._upload_dataset import DEFAULT_SIMULTANEOUS_UPLOADS
from wkcuber.api.dataset import Dataset

from .utils import add_verbose_flag, setup_logging


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "source_path", help="Directory containing the source WKW dataset.", type=Path
    )

    parser.add_argument(
        "--token",
        help="Auth token of the user on webKnossos",
        default=None,
    )

    parser.add_argument(
        "--url", help="Base url of the webKnossos instance", default=None
    )

    parser.add_argument(
        "--jobs",
        "-j",
        default=DEFAULT_SIMULTANEOUS_UPLOADS,
        type=int,
        help=f"Number of simultaneous upload processes. Defaults to {DEFAULT_SIMULTANEOUS_UPLOADS}.",
    )

    parser.add_argument(
        "--name",
        help="Specify a new name for the dataset. Defaults to the name specified in `datasource-properties.json`.",
        default=None,
    )

    add_verbose_flag(parser)

    return parser


def upload_dataset(
    source_path: Path, url: str, token: Optional[str], jobs: int, name: Optional[str]
) -> None:
    with webknossos_context(url=url, token=token):
        Dataset.open(source_path).upload(new_dataset_name=name, jobs=jobs)


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)
    url = (
        args.url
        if args.url is not None
        else environ.get("WK_URL", DEFAULT_WEBKNOSSOS_URL)
    )
    token = args.token if args.token is not None else environ.get("WK_TOKEN", None)
    assert (
        token is not None
    ), f"An auth token needs to be supplied either through the --token command line arg or the WK_TOKEN environment variable. Retrieve your auth token on {url}/auth/token."

    upload_dataset(args.source_path, url, token, args.jobs, args.name)



from typing import List, Dict

from wkcuber import downsample_mags
from .compress import compress_mag_inplace
from .metadata import refresh_metadata
from .utils import (
    add_distribution_flags,
    add_isotropic_flag,
    setup_logging,
    add_sampling_mode_flag,
)
from .mag import Mag
from .converter import (
    create_parser as create_conversion_parser,
    main as auto_detect_and_run_conversion,
)
from .versioning import get_available_version
from argparse import Namespace, ArgumentParser
from pathlib import Path


def detect_present_mags(target_path: Path) -> Dict[Path, List[Mag]]:
    layer_path_to_mags: Dict[Path, List[Mag]] = dict()
    layer_paths = list([p for p in target_path.iterdir() if p.is_dir()])
    for layer_p in layer_paths:
        layer_path_to_mags.setdefault(layer_p, list())
        mag_paths = list([p for p in layer_p.iterdir() if p.is_dir()])
        for mag_p in mag_paths:
            try:
                mag = Mag(mag_p.name)
            except (AssertionError, ValueError) as _:
                continue
            layer_path_to_mags[layer_p].append(mag)

    return layer_path_to_mags


def create_parser() -> ArgumentParser:
    parser = create_conversion_parser()

    parser.add_argument(
        "--max_mag",
        "-m",
        help="Max resolution to be downsampled. Needs to be a power of 2. In case of anisotropic downsampling, "
        "the process is considered done when max(current_mag) >= max(max_mag) where max takes the "
        "largest dimension of the mag tuple x, y, z. For example, a maximum mag value of 8 (or 8-8-8) "
        "will stop the downsampling as soon as a magnification is produced for which one dimension is "
        "equal or larger than 8. "
        "The default value is calculated depending on the dataset size. In the lowest Mag, the size will be "
        "smaller than 100vx per dimension",
        type=int,
        default=None,
    )

    parser.add_argument(
        "--no_compress",
        help="Don't compress this data",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--version", action="version", version=f"wkcuber {get_available_version()}"
    )

    parser.add_argument("--name", "-n", help="Name of the dataset", default=None)
    add_isotropic_flag(parser)
    add_sampling_mode_flag(parser)
    add_distribution_flags(parser)

    return parser


def cube_with_args(args: Namespace) -> None:
    if args.isotropic is not None:
        raise DeprecationWarning(
            "The flag 'isotropic' is deprecated. Consider using '--sampling_mode isotropic' instead."
        )

    auto_detect_and_run_conversion(args)

    layer_path_to_mags: Dict[Path, List[Mag]] = detect_present_mags(args.target_path)

    if not args.no_compress:
        for (layer_path, mags) in layer_path_to_mags.items():
            layer_name = layer_path.name
            for mag in mags:
                compress_mag_inplace(args.target_path, layer_name, mag, args)

    for (layer_path, mags) in layer_path_to_mags.items():
        layer_name = layer_path.name
        mags.sort()
        downsample_mags(
            path=args.target_path,
            layer_name=layer_name,
            from_mag=mags[-1],
            max_mag=None if args.max_mag is None else Mag(args.max_mag),
            interpolation_mode="default",
            compress=not args.no_compress,
            sampling_mode=args.sampling_mode,
            args=args,
        )

    refresh_metadata(args.target_path)


def main() -> None:
    args = create_parser().parse_args()
    setup_logging(args)
    cube_with_args(args)


if __name__ == "__main__":
    main()



"""
# wkcuber

This module provides CLI tools to modify or interact with datasets.
For information regarding the other functionalities, refer to the `README.md`.
"""

from .api.dataset import Dataset
from .cubing import cubing
from .downsampling import downsample_mags
from .compress import compress_mag
from .mag import Mag
from .metadata import write_webknossos_metadata



import argparse
from functools import partial
import logging
from pathlib import Path
from typing import Optional, Tuple, Union
import numpy as np

from webknossos.dataset.defaults import DEFAULT_WKW_FILE_LEN
from webknossos import BoundingBox, Dataset, Mag, MagView
from webknossos.utils import (
    get_executor_for_args,
    time_start,
    time_stop,
    wait_and_ensure_success,
)
from .utils import (
    add_distribution_flags,
    add_interpolation_flag,
    add_sampling_mode_flag,
    add_scale_flag,
    add_verbose_flag,
    DEFAULT_WKW_VOXELS_PER_BLOCK,
    get_executor_args,
    setup_logging,
)


logger = logging.getLogger(__name__)


def parse_shape(shape: str) -> Tuple[float, ...]:
    try:
        return tuple(int(x) for x in shape.split(","))
    except Exception as e:
        raise argparse.ArgumentTypeError("The shape could not be parsed") from e


def parse_flip_axes(flip_axes: str) -> Tuple[int, ...]:
    try:
        indices = tuple(int(x) for x in flip_axes.split(","))
    except Exception as e:
        raise argparse.ArgumentTypeError("The flip_axes could not be parsed") from e
    if [i for i in indices if i < 0 or i > 3]:
        raise argparse.ArgumentTypeError("The flip_axes contains out-of-bound values")
    return indices


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "source_path",
        help="Path to raw file to convert",
        type=Path,
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated WKW dataset.", type=Path
    )

    parser.add_argument(
        "--input_dtype",
        help="Input dataset datatype (e.g. uint8, uint16, float32).",
        required=True,
    )

    parser.add_argument(
        "--shape",
        help="Shape of the dataset (width, height, depth)",
        type=parse_shape,
        required=True,
    )

    add_scale_flag(parser, required=False)

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation).",
        default="color",
    )

    parser.add_argument(
        "--order",
        help="The input data storage layout:"
        "either 'F' for Fortran-style/column-major order (the default), "
        "or 'C' for C-style/row-major order. "
        "Note: Axes are expected in  (x, y, z) order.",
        choices=("C", "F"),
        default="F",
    )

    parser.add_argument(
        "--flip_axes",
        help="The axes which should be flipped. "
        "Input format is a comma separated list of axis indices. "
        "For example, 1,2,3 will flip the x, y and z axes.",
        default=None,
        type=parse_flip_axes,
    )

    parser.add_argument(
        "--no_compress",
        help="Don't compress saved data",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--max_mag",
        "-m",
        help="Max resolution to be downsampled. Needs to be a power of 2. In case of anisotropic downsampling, "
        "the process is considered done when max(current_mag) >= max(max_mag) where max takes the "
        "largest dimension of the mag tuple x, y, z. For example, a maximum mag value of 8 (or 8-8-8) "
        "will stop the downsampling as soon as a magnification is produced for which one dimension is "
        "equal or larger than 8. "
        "The default value is calculated depending on the dataset size. In the lowest Mag, the size will be "
        "smaller than 100vx per dimension",
        type=int,
        default=None,
    )

    add_sampling_mode_flag(parser)
    add_interpolation_flag(parser)

    add_distribution_flags(parser)

    add_verbose_flag(parser)

    return parser


def _raw_chunk_converter(
    bounding_box: BoundingBox,
    source_raw_path: Path,
    target_mag_view: MagView,
    input_dtype: str,
    shape: Tuple[int, int, int],
    order: str,
    flip_axes: Optional[Union[int, Tuple[int, ...]]],
) -> None:
    logging.info(f"Conversion of {bounding_box.topleft}")
    source_data = np.memmap(
        source_raw_path, dtype=input_dtype, mode="r", shape=(1,) + shape, order=order
    )

    if flip_axes:
        source_data = np.flip(source_data, flip_axes)

    contiguous_chunk = source_data[(slice(None),) + bounding_box.to_slices()].copy(
        order="F"
    )
    target_mag_view.write(contiguous_chunk, bounding_box.topleft)


def convert_raw(
    source_raw_path: Path,
    target_path: Path,
    layer_name: str,
    input_dtype: str,
    shape: Tuple[int, int, int],
    order: str = "F",
    scale: Optional[Tuple[float, float, float]] = (1.0, 1.0, 1.0),
    flip_axes: Optional[Union[int, Tuple[int, ...]]] = None,
    compress: bool = True,
    file_len: int = DEFAULT_WKW_FILE_LEN,
    executor_args: Optional[argparse.Namespace] = None,
) -> MagView:
    assert order in ("C", "F")
    time_start(f"Conversion of {source_raw_path}")

    if scale is None:
        scale = 1.0, 1.0, 1.0
    wk_ds = Dataset(target_path, scale=scale, exist_ok=True)
    wk_layer = wk_ds.get_or_add_layer(
        layer_name,
        "color",
        dtype_per_layer=np.dtype(input_dtype),
        num_channels=1,
    )
    wk_layer.bounding_box = BoundingBox((0, 0, 0), shape)
    wk_mag = wk_layer.get_or_add_mag("1", file_len=file_len, compress=compress)

    # Parallel chunk conversion
    with get_executor_for_args(executor_args) as executor:
        wait_and_ensure_success(
            executor.map_to_futures(
                partial(
                    _raw_chunk_converter,
                    source_raw_path=source_raw_path,
                    target_mag_view=wk_mag,
                    input_dtype=input_dtype,
                    shape=shape,
                    order=order,
                    flip_axes=flip_axes,
                ),
                wk_layer.bounding_box.chunk(
                    chunk_size=(DEFAULT_WKW_VOXELS_PER_BLOCK * file_len,) * 3
                ),
            )
        )

    time_stop(f"Conversion of {source_raw_path}")
    return wk_mag


def main(args: argparse.Namespace) -> None:
    source_path = args.source_path

    if source_path.is_dir():
        logger.error("source_path is not a file")
        return

    executor_args = get_executor_args(args)

    mag_view = convert_raw(
        source_path,
        args.target_path,
        args.layer_name,
        args.input_dtype,
        args.shape,
        args.order,
        args.scale,
        args.flip_axes,
        not args.no_compress,
        executor_args=executor_args,
    )

    mag_view.layer.downsample(
        from_mag=mag_view.mag,
        max_mag=None if args.max_mag is None else Mag(args.max_mag),
        interpolation_mode=args.interpolation_mode,
        compress=not args.no_compress,
        sampling_mode=args.sampling_mode,
        args=executor_args,
    )


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    main(args)



import logging
from pathlib import Path
from typing import Tuple, cast, Optional
from webknossos.utils import time_start, time_stop

import wkw
from argparse import ArgumentParser, Namespace

from .utils import (
    add_verbose_flag,
    open_wkw,
    open_knossos,
    WkwDatasetInfo,
    KnossosDatasetInfo,
    ensure_wkw,
    add_distribution_flags,
    get_executor_for_args,
    wait_and_ensure_success,
    setup_logging,
)
from .knossos import CUBE_EDGE_LEN
from .metadata import convert_element_class_to_dtype


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "source_path",
        help="Directory containing the source KNOSSOS dataset.",
        type=Path,
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated WKW dataset.", type=Path
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument("--mag", "-m", help="Magnification level", type=int, default=1)

    add_verbose_flag(parser)
    add_distribution_flags(parser)

    return parser


def convert_cube_job(
    args: Tuple[Tuple[int, int, int], KnossosDatasetInfo, WkwDatasetInfo]
) -> None:
    cube_xyz, source_knossos_info, target_wkw_info = args
    offset = cast(Tuple[int, int, int], tuple(x * CUBE_EDGE_LEN for x in cube_xyz))
    size = cast(Tuple[int, int, int], (CUBE_EDGE_LEN,) * 3)
    time_start(f"Converting of {offset}")

    with open_knossos(source_knossos_info) as source_knossos, open_wkw(
        target_wkw_info
    ) as target_wkw:
        cube_data = source_knossos.read(offset, size)
        target_wkw.write(offset, cube_data)

    time_stop(f"Converting of {offset}")


def convert_knossos(
    source_path: Path,
    target_path: Path,
    layer_name: str,
    dtype: str,
    mag: int = 1,
    args: Optional[Namespace] = None,
) -> None:
    source_knossos_info = KnossosDatasetInfo(source_path, dtype)
    target_wkw_info = WkwDatasetInfo(
        target_path, layer_name, mag, wkw.Header(convert_element_class_to_dtype(dtype))
    )

    ensure_wkw(target_wkw_info)

    with open_knossos(source_knossos_info) as source_knossos:
        with get_executor_for_args(args) as executor:
            knossos_cubes = list(source_knossos.list_cubes())
            if len(knossos_cubes) == 0:
                logging.error(
                    "No input KNOSSOS cubes found. Make sure to pass the path which points to a KNOSSOS magnification (e.g., testdata/knossos/color/1)."
                )
                exit(1)

            knossos_cubes.sort()
            job_args = []
            for cube_xyz in knossos_cubes:
                job_args.append((cube_xyz, source_knossos_info, target_wkw_info))

            wait_and_ensure_success(
                executor.map_to_futures(convert_cube_job, job_args),
                f"Converting knossos layer {layer_name}",
            )


def main(args: Namespace) -> None:
    convert_knossos(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        args.mag,
        args,
    )


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    main(args)



import logging
from typing import List, Tuple, Optional, cast, Any

import numpy as np
from argparse import ArgumentParser, Namespace
from os import path
from pathlib import Path
from natsort import natsorted

from webknossos.dataset import (
    Dataset,
    COLOR_CATEGORY,
    SEGMENTATION_CATEGORY,
    View,
    SegmentationLayer,
    Layer,
)
from webknossos.geometry import BoundingBox, Vec3Int
from webknossos.utils import time_start, time_stop
from .mag import Mag
from .downsampling_utils import (
    parse_interpolation_mode,
    downsample_unpadded_data,
    InterpolationModes,
)
from .utils import (
    get_chunks,
    find_files,
    add_batch_size_flag,
    add_verbose_flag,
    add_distribution_flags,
    add_interpolation_flag,
    get_executor_for_args,
    wait_and_ensure_success,
    setup_logging,
    add_scale_flag,
)
from .image_readers import image_reader

BLOCK_LEN = 32


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "source_path", help="Directory containing the input images.", type=Path
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated dataset.", type=Path
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--skip_first_z_slices",
        help="The number of z slices to skip at the beginning. This is useful to continue at a specific z, if a previous run was interrupted.",
        default=0,
        type=int,
    )

    parser.add_argument(
        "--start_z",
        help="The offset of the first z slice.",
        default=0,
        type=int,
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default=None,
    )

    parser.add_argument(
        "--wkw_file_len",
        default=32,
        type=int,
        help="Amount of blocks which are written per dimension to a wkw cube. The default value of 32 means that 1024 slices are written to one cube (since one block has 32**3 voxels by default). For single-channel uint8 data, this results in 1 GB per cube file. If file_len is set to 1, only 32 slices are written to one cube. Must be a power of two.",
    )

    add_batch_size_flag(parser)

    parser.add_argument(
        "--pad",
        help="Automatically pad image files at the bottom and right borders. "
        "Use this, when the input images don't have a common size, but have "
        "their origin at (0, 0).",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--target_mag",
        help="Automatically downsamples the cubed images to the provided "
        "magnification before writing to disk. The magnification can "
        "be provided like 2-2-1.",
        default="1",
        type=Mag,
    )

    parser.add_argument(
        "--channel_index",
        "-c",
        type=int,
        default=None,
        help="Select a single channel to be cubed into a layer. This can be helpful if several channels should be converted to multiple data layers. Note that webKnossos only supports multiple channels if these are three uint8 channels. If this is not the case, use --channel_index to create a layer per channel.",
    )

    parser.add_argument(
        "--sample_index",
        type=int,
        default=None,
        help="Select a single sample of a specific channel to be cubed into a layer. This option is only valid if channel_index is set. Since webKnossos only supports multiple uint8 channels, it may be necessary to cube a multi-sample dataset to different layers.",
    )

    add_scale_flag(parser)
    add_interpolation_flag(parser)
    add_verbose_flag(parser)
    add_distribution_flags(parser)

    return parser


def find_source_filenames(source_path: Path) -> List[Path]:
    # Find all source files that have a matching file extension
    if source_path.is_dir():
        source_path_str = path.join(source_path, "*")
    else:
        source_path_str = str(source_path)

    source_files = list(find_files(source_path_str, image_reader.readers.keys()))

    assert len(source_files) > 0, (
        "No image files found in path "
        + source_path_str
        + ". Supported suffixes are "
        + str(image_reader.readers.keys())
        + "."
    )

    return natsorted(source_files)


def read_image_file(
    file_name: Path,
    dtype: type,
    z_slice: int,
    channel_index: Optional[int],
    sample_index: Optional[int],
) -> np.ndarray:
    try:
        return image_reader.read_array(
            file_name, dtype, z_slice, channel_index, sample_index
        )
    except Exception as exc:
        logging.error("Reading of file={} failed with {}".format(file_name, exc))
        raise exc


def cubing_job(
    args: Tuple[
        View,
        Mag,
        InterpolationModes,
        List[str],
        int,
        bool,
        Optional[int],
        Optional[int],
        str,
        int,
    ]
) -> Any:
    (
        target_view,
        target_mag,
        interpolation_mode,
        source_file_batches,
        batch_size,
        pad,
        channel_index,
        sample_index,
        dtype,
        num_channels,
    ) = args

    downsampling_needed = target_mag != Mag(1)
    largest_value_in_chunk = 0  # This is used to compute the largest_segmentation_id if it is a segmentation layer

    max_image_size = (target_view.size[0], target_view.size[1])

    # Iterate over batches of continuous z sections
    # The batches have a maximum size of `batch_size`
    # Batched iterations allows to utilize IO more efficiently
    first_z_idx = target_view.global_offset.z
    for source_file_batch in get_chunks(source_file_batches, batch_size):
        try:
            time_start(
                f"Cubing of z={first_z_idx}-{first_z_idx + len(source_file_batch)}"
            )
            # Allocate a large buffer for all images in this batch
            # Shape will be (channel_count, x, y, z)
            # Using fortran order for the buffer, prevents that the data has to be copied in rust
            buffer_shape = (
                [num_channels] + list(max_image_size) + [len(source_file_batch)]
            )
            buffer = np.empty(buffer_shape, dtype=dtype, order="F")

            # Iterate over each z section in the batch
            for i, file_name in enumerate(source_file_batch):
                z = first_z_idx + i
                # Image shape will be (x, y, channel_count, z=1)
                image = read_image_file(
                    file_name,
                    target_view.header.voxel_type,
                    z,
                    channel_index,
                    sample_index,
                )

                if pad:
                    image = np.pad(
                        image,
                        mode="constant",
                        pad_width=[
                            (0, max_image_size[0] - image.shape[0]),
                            (0, max_image_size[1] - image.shape[1]),
                            (0, 0),
                            (0, 0),
                        ],
                    )
                else:
                    assert (
                        image.shape[0:2] == max_image_size
                    ), "Section z={} has the wrong dimensions: {} (expected {}). Consider using --pad.".format(
                        z, image.shape, max_image_size
                    )
                buffer[:, :, :, i] = image.transpose((2, 0, 1, 3))[:, :, :, 0]
            del image

            if downsampling_needed:
                buffer = downsample_unpadded_data(
                    buffer, target_mag, interpolation_mode
                )
            buffer_z_offset = (
                first_z_idx - target_view.global_offset.z
            ) // target_mag.z
            target_view.write(offset=(0, 0, buffer_z_offset), data=buffer)
            largest_value_in_chunk = max(largest_value_in_chunk, np.max(buffer))
            time_stop(
                f"Cubing of z={first_z_idx}-{first_z_idx + len(source_file_batch)}"
            )
            first_z_idx += len(source_file_batch)

        except Exception as exc:
            logging.error(
                "Cubing of z={}-{} failed with {}".format(
                    first_z_idx, first_z_idx + len(source_file_batch), exc
                )
            )
            raise exc

    return largest_value_in_chunk


def get_channel_and_sample_count_and_dtype(source_path: Path) -> Tuple[int, int, str]:
    source_files = find_source_filenames(source_path)
    assert (
        len(source_files) > 0
    ), f"Failed to detect channel count and dtype. No sources found in {source_path}"
    return (
        image_reader.read_channel_count(source_files[0]),
        image_reader.read_sample_count(source_files[0]),
        image_reader.read_dtype(source_files[0]),
    )


def cubing(
    source_path: Path,
    target_path: Path,
    layer_name: str,
    batch_size: Optional[int],
    channel_index: Optional[int],
    sample_index: Optional[int],
    dtype: Optional[str],
    target_mag_str: str,
    wkw_file_len: int,
    interpolation_mode_str: str,
    start_z: int,
    skip_first_z_slices: int,
    pad: bool,
    scale: Tuple[float, float, float],
    executor_args: Namespace,
) -> Layer:
    source_files = find_source_filenames(source_path)

    all_num_x, all_num_y = zip(
        *[
            image_reader.read_dimensions(source_files[i])
            for i in range(len(source_files))
        ]
    )
    num_x = max(all_num_x)
    num_y = max(all_num_y)
    # All images are assumed to have equal channels and samples
    num_channels = image_reader.read_channel_count(source_files[0])
    num_samples = image_reader.read_sample_count(source_files[0])
    num_output_channels = num_channels * num_samples
    if channel_index is not None:
        # if there is no c axis, but someone meant to only use one channel/sample, set the sample index instead
        if sample_index is None and num_channels == 1 and channel_index > 0:
            sample_index = channel_index
            channel_index = 0

        assert (
            0 <= channel_index < num_channels
        ), "Selected channel is invalid. Please check the number of channels in the source file."
        num_output_channels = num_samples
    if sample_index is not None:
        # if no channel axis exists, it is valid to only set the sample index. Set channel index to 0 to avoid confusion
        if channel_index is None and num_channels == 1:
            channel_index = 0
        assert (
            channel_index is not None
        ), "Sample index is only valid if a channel index is also set."
        assert (
            0 <= sample_index < num_samples
        ), "Selected sample is invalid. Please check the number of samples in the source file."
        num_output_channels = 1
    num_z_slices_per_file = image_reader.read_z_slices_per_file(source_files[0])
    assert (
        num_z_slices_per_file == 1 or len(source_files) == 1
    ), "Multi page TIFF support only for single files"
    if num_z_slices_per_file > 1:
        num_z = num_z_slices_per_file
    else:
        num_z = len(source_files)

    if dtype is None:
        dtype = image_reader.read_dtype(source_files[0])

    if batch_size is None:
        batch_size = BLOCK_LEN

    target_mag = Mag(target_mag_str)

    target_ds = Dataset(target_path, scale=scale, exist_ok=True)
    is_segmentation_layer = layer_name == "segmentation"

    if is_segmentation_layer:
        target_layer = target_ds.get_or_add_layer(
            layer_name,
            SEGMENTATION_CATEGORY,
            dtype_per_channel=dtype,
            num_channels=num_output_channels,
            largest_segment_id=0,
        )
    else:
        target_layer = target_ds.get_or_add_layer(
            layer_name,
            COLOR_CATEGORY,
            dtype_per_channel=dtype,
            num_channels=num_output_channels,
        )
    target_layer.bounding_box = target_layer.bounding_box.extended_by(
        BoundingBox(
            Vec3Int(0, 0, start_z + skip_first_z_slices) * target_mag,
            Vec3Int(num_x, num_y, num_z - skip_first_z_slices) * target_mag,
        )
    )

    target_mag_view = target_layer.get_or_add_mag(
        target_mag, file_len=wkw_file_len, block_len=BLOCK_LEN
    )

    interpolation_mode = parse_interpolation_mode(
        interpolation_mode_str, target_layer.category
    )
    if target_mag != Mag(1):
        logging.info(
            f"Downsampling the cubed image to {target_mag} in memory with interpolation mode {interpolation_mode}."
        )

    logging.info("Found source files: count={} size={}x{}".format(num_z, num_x, num_y))

    with get_executor_for_args(executor_args) as executor:
        job_args = []
        # We iterate over all z sections
        for z in range(skip_first_z_slices, num_z, BLOCK_LEN):
            # The z is used to access the source files. However, when writing the data, the `start_z` has to be considered.
            max_z = min(num_z, z + BLOCK_LEN)
            # Prepare source files array
            if len(source_files) > 1:
                source_files_array = source_files[z:max_z]
            else:
                source_files_array = source_files * (max_z - z)

            # Prepare job
            job_args.append(
                (
                    target_mag_view.get_view(
                        (0, 0, z + start_z),
                        (num_x, num_y, max_z - z),
                    ),
                    target_mag,
                    interpolation_mode,
                    source_files_array,
                    batch_size,
                    pad,
                    channel_index,
                    sample_index,
                    dtype,
                    target_layer.num_channels,
                )
            )

        largest_segment_id_per_chunk = wait_and_ensure_success(
            executor.map_to_futures(cubing_job, job_args),
            progress_desc=f"Cubing from {skip_first_z_slices} to {num_z}",
        )
        if is_segmentation_layer:
            largest_segment_id = max(largest_segment_id_per_chunk)
            cast(
                SegmentationLayer, target_layer
            ).largest_segment_id = largest_segment_id

    # Return layer
    return target_layer


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    arg_dict = vars(args)

    cubing(
        args.source_path,
        args.target_path,
        args.layer_name,
        arg_dict.get("batch_size"),
        arg_dict.get("channel_index"),
        arg_dict.get("sample_index"),
        arg_dict.get("dtype"),
        args.target_mag,
        args.wkw_file_len,
        args.interpolation_mode,
        args.start_z,
        args.skip_first_z_slices,
        args.pad,
        args.scale,
        args,
    )



import logging
from pathlib import Path
from typing import List, Tuple

import wkw
import numpy as np
from argparse import ArgumentParser
from itertools import product

from .metadata import detect_bbox

from .utils import (
    add_verbose_flag,
    open_wkw,
    WkwDatasetInfo,
    ensure_wkw,
    add_distribution_flags,
    setup_logging,
    get_executor_for_args,
    wait_and_ensure_success,
)


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "source_path", help="Directory containing the datasource properties.", type=Path
    )

    parser.add_argument(
        "target_path", help="Output directory for the generated dataset.", type=Path
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--dtype",
        "-d",
        help="Target datatype (e.g. uint8, uint16, uint32)",
        default="uint8",
    )

    parser.add_argument(
        "--wkw_file_len", help="Target file length (default 32)", type=int, default=32
    )

    parser.add_argument(
        "--no_compression",
        help="Use compression, default false",
        type=bool,
        default=False,
    )

    add_verbose_flag(parser)
    add_distribution_flags(parser)

    return parser


def next_lower_divisible_by(number: int, divisor: int) -> int:
    remainder = number % divisor
    return number - remainder


def next_higher_divisible_by(number: int, divisor: int) -> int:
    remainder = number % divisor
    return number - remainder + divisor


def recube(
    source_path: Path,
    target_path: Path,
    layer_name: str,
    dtype: str,
    wkw_file_len: int = 32,
    compression: bool = True,
) -> None:
    if compression:
        block_type = wkw.Header.BLOCK_TYPE_LZ4
    else:
        block_type = wkw.Header.BLOCK_TYPE_RAW

    target_wkw_header = wkw.Header(
        np.dtype(dtype), file_len=wkw_file_len, block_type=block_type
    )
    target_wkw_info = WkwDatasetInfo(target_path, layer_name, 1, target_wkw_header)
    source_wkw_header = wkw.Header(np.dtype(dtype))
    source_wkw_info = WkwDatasetInfo(source_path, layer_name, 1, source_wkw_header)

    ensure_wkw(target_wkw_info)

    bounding_box_dict = detect_bbox(source_wkw_info.dataset_path, layer_name)
    if bounding_box_dict is None:
        raise ValueError("Failed to detect bounding box.")

    bounding_box = (
        bounding_box_dict["topLeft"],
        [
            bounding_box_dict["width"],
            bounding_box_dict["height"],
            bounding_box_dict["depth"],
        ],
    )
    bottom_right = [
        coord + size for coord, size in zip(bounding_box[0], bounding_box[1])
    ]

    wkw_cube_size = wkw_file_len * target_wkw_header.block_len

    outer_bounding_box_tl = list(
        map(lambda lx: next_lower_divisible_by(lx, wkw_cube_size), bounding_box[0])
    )
    outer_bounding_box_br = list(
        map(lambda lx: next_higher_divisible_by(lx, wkw_cube_size), bottom_right)
    )
    outer_bounding_box_size = [
        outer_bounding_box_br[0] - outer_bounding_box_tl[0],
        outer_bounding_box_br[1] - outer_bounding_box_tl[1],
        outer_bounding_box_br[2] - outer_bounding_box_tl[2],
    ]

    target_cube_addresses = product(
        range(0, outer_bounding_box_size[0], wkw_cube_size),
        range(0, outer_bounding_box_size[1], wkw_cube_size),
        range(0, outer_bounding_box_size[2], wkw_cube_size),
    )

    with get_executor_for_args(args) as executor:
        job_args = []
        for target_cube_xyz in target_cube_addresses:
            job_args.append(
                (
                    source_wkw_info,
                    target_wkw_info,
                    outer_bounding_box_size,
                    outer_bounding_box_tl,
                    wkw_cube_size,
                    target_cube_xyz,
                )
            )
        wait_and_ensure_success(executor.map_to_futures(recubing_cube_job, job_args))

    logging.info(f"{layer_name} successfully resampled!")


def recubing_cube_job(
    args: Tuple[
        WkwDatasetInfo, WkwDatasetInfo, List[int], List[int], int, Tuple[int, int, int]
    ]
) -> None:
    (
        source_wkw_info,
        target_wkw_info,
        _outer_bounding_box_size,
        outer_bounding_box_tl,
        wkw_cube_size,
        target_cube_xyz,
    ) = args

    with open_wkw(source_wkw_info) as source_wkw_dataset:
        with open_wkw(target_wkw_info) as target_wkw_dataset:
            top_left = [
                outer_bounding_box_tl[0] + target_cube_xyz[0],
                outer_bounding_box_tl[1] + target_cube_xyz[1],
                outer_bounding_box_tl[2] + target_cube_xyz[2],
            ]

            logging.debug("Writing at {}".format(top_left))

            data_cube = source_wkw_dataset.read(
                top_left, (wkw_cube_size, wkw_cube_size, wkw_cube_size)
            )

            target_wkw_dataset.write(top_left, data_cube)


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)

    recube(
        args.source_path,
        args.target_path,
        args.layer_name,
        args.dtype,
        args.wkw_file_len,
        not args.no_compression,
    )



from pathlib import Path

import shutil
import logging
from argparse import ArgumentParser, Namespace
from typing import Optional
from os import makedirs

from wkcuber.api.dataset import Dataset
from .mag import Mag

from .utils import add_verbose_flag, add_distribution_flags, setup_logging
from typing import List

BACKUP_EXT = ".bak"


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()

    parser.add_argument(
        "source_path", help="Directory containing the source WKW dataset.", type=Path
    )

    parser.add_argument(
        "target_path",
        help="Output directory for the compressed WKW dataset.",
        nargs="?",
        default=None,
        type=Path,
    )

    parser.add_argument(
        "--layer_name",
        "-l",
        help="Name of the cubed layer (color or segmentation)",
        default="color",
    )

    parser.add_argument(
        "--mag", "-m", nargs="*", help="Magnification level", default=None
    )

    add_verbose_flag(parser)
    add_distribution_flags(parser)

    return parser


def compress_mag(
    source_path: Path,
    layer_name: str,
    target_path: Path,
    mag: Mag,
    args: Optional[Namespace] = None,
) -> None:
    Dataset.open(source_path).get_layer(layer_name).get_mag(mag).compress(
        target_path=Path(target_path), args=args
    )


def compress_mag_inplace(
    target_path: Path, layer_name: str, mag: Mag, args: Optional[Namespace] = None
) -> None:
    Dataset.open(target_path).get_layer(layer_name).get_mag(mag).compress(args=args)


def compress_mags(
    source_path: Path,
    layer_name: str,
    target_path: Optional[Path] = None,
    mags: Optional[List[Mag]] = None,
    args: Optional[Namespace] = None,
) -> None:
    if target_path is None:
        target = source_path.with_suffix(".tmp")
    else:
        target = target_path

    layer = Dataset.open(source_path).get_layer(layer_name)
    if mags is None:
        mags = list(layer.mags.keys())

    for mag, mag_view in Dataset.open(source_path).get_layer(layer_name).mags.items():
        if mag in mags:
            mag_view.compress(target_path=Path(target), args=args)

    if target_path is None:
        backup_dir = source_path.with_suffix(BACKUP_EXT)
        makedirs(backup_dir / layer_name, exist_ok=True)
        for mag in mags:
            shutil.move(
                str(source_path / layer_name / str(mag)),
                str(backup_dir / layer_name / str(mag)),
            )
            shutil.move(
                str(target / layer_name / str(mag)),
                str(source_path / layer_name / str(mag)),
            )
        shutil.rmtree(target)
        logging.info(
            "Old files are still present in '{0}.bak'. Please remove them when not required anymore.".format(
                source_path
            )
        )


if __name__ == "__main__":
    args = create_parser().parse_args()
    setup_logging(args)
    compress_mags(args.source_path, args.layer_name, args.target_path, args.mag, args)
